/*	$NetBSD: scsipiconf.h,v 1.108 2006/11/26 05:01:09 itohy Exp $	*/

/*-
 * Copyright (c) 1998, 1999, 2000, 2004 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Charles M. Hannum; by Jason R. Thorpe of the Numerical Aerospace
 * Simulation Facility, NASA Ames Research Center.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *        This product includes software developed by the NetBSD
 *        Foundation, Inc. and its contributors.
 * 4. Neither the name of The NetBSD Foundation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Originally written by Julian Elischer (julian@tfs.com)
 * for TRW Financial Systems for use under the MACH(2.5) operating system.
 *
 * TRW Financial Systems, in accordance with their agreement with Carnegie
 * Mellon University, makes this software available to CMU to distribute
 * or use in any manner that they see fit as long as this message is kept with
 * the software. For this reason TFS also grants any other persons or
 * organisations permission to use or modify this software.
 *
 * TFS supplies this software to be publicly redistributed
 * on the understanding that TFS is not responsible for the correct
 * functioning of this software in any circumstances.
 *
 * Ported to run under 386BSD by Julian Elischer (julian@tfs.com) Sept 1992
 *
 * 2009: modified by minoru murashima.
 */

#ifndef _DEV_SCSIPI_SCSIPICONF_H_
#define _DEV_SCSIPI_SCSIPICONF_H_

#include <sys/queue.h>
#include <sys/callout.h>
#include <sys/buf.h>
#include <scsipi/scsi_spc.h>
#include <scsipi/sdvar.h>

/*
 * Number of tag words in a periph structure:
 *
 *	n_tag_words = ((256 / NBBY) / sizeof(u_int32_t))
 */
#define	PERIPH_NTAGWORDS	((256 / 8) / sizeof(u_int32_t))

/*
 * scsipi_async_event_t:
 *
 *	Asynchronous events from the adapter to the mid-layer and
 *	peripherial.
 *
 *	Arguments:
 *
 *	ASYNC_EVENT_MAX_OPENINGS	scsipi_max_openings * -- max
 *								openings, device specified in
 *								parameters
 *
 *	ASYNC_EVENT_XFER_MODE		scsipi_xfer_mode * -- xfer mode
 *								parameters changed for I_T Nexus
 *	ASYNC_EVENT_RESET			NULL - channel has been reset
 */
typedef enum {
	ASYNC_EVENT_MAX_OPENINGS,	/* set max openings on periph */
	ASYNC_EVENT_XFER_MODE,		/* xfer mode update for I_T */
	ASYNC_EVENT_RESET			/* channel reset */
} scsipi_async_event_t;

/*
 * scsipi_adapter_req_t:
 *
 *	Requests that can be made of an adapter.
 *
 *	Arguments:
 *
 *	ADAPTER_REQ_RUN_XFER		scsipi_xfer * -- the xfer which
 *					is to be run
 *
 *	ADAPTER_REQ_GROW_RESOURCES	no argument
 *
 *	ADAPTER_REQ_SET_XFER_MODE	scsipi_xfer_mode * -- set the xfer
 *					mode for the I_T Nexus according to
 *					this
 */
typedef enum {
	ADAPTER_REQ_RUN_XFER,		/* run a scsipi_xfer */
	ADAPTER_REQ_GROW_RESOURCES,	/* grow xfer execution resources */
	ADAPTER_REQ_SET_XFER_MODE	/* set xfer mode */
} scsipi_adapter_req_t;

struct scsipi_generic {
	u_int8_t opcode;
	u_int8_t bytes[15];
};

typedef	int	boolean;

/*
 * This describes matching information for scsipi_inqmatch().  The more things
 * match, the higher the configuration priority.
 */
struct scsipi_inquiry_pattern {
	u_int8_t type;
	boolean removable;
	const char *vendor;
	const char *product;
	const char *revision;
};

struct scsipi_xfer;
struct scsipi_periph;
struct scsipi_channel;

/*
 * scsipi_adapter:
 *
 *	This structure describes an instance of a SCSIPI adapter.
 *
 *	Note that `adapt_openings' is used by (the common case of) adapters
 *	which have per-adapter resources.  If an adapter's command resources
 *	are associated with a channel, then the `chan_openings' below will
 *	be used instead.
 *
 *	Note that all adapter entry points take a pointer to a channel,
 *	as an adapter may have more than one channel, and the channel
 *	structure contains the channel number.
 */
struct scsipi_adapter {
	struct device *adapt_dev;	/* pointer to adapter's device */
	int		adapt_nchannels;	/* number of adapter channels */
	int		adapt_refcnt;		/* adapter's reference count */
	int		adapt_openings;		/* total # of command openings */
	int		adapt_max_periph;	/* max openings per periph */
	int		adapt_flags;

	void	(*adapt_request)(struct scsipi_channel *, scsipi_adapter_req_t, void *);
	void	(*adapt_minphys)(struct buf *);
	int		(*adapt_ioctl)(struct scsipi_channel *, u_long, caddr_t, int, struct proc *);
	int		(*adapt_enable)(struct device *, int);
	int		(*adapt_getgeom)(struct scsipi_periph *, struct disk_parms *, u_long);
	int		(*adapt_accesschk)(struct scsipi_periph *, struct scsipi_inquiry_pattern *);
};

/*
 * scsipi_periphsw:
 *
 *	Callbacks into periph driver from midlayer.
 *
 *	psw_error	Called by the bustype's interpret-sense routine
 *			to do periph-specific sense handling.
 *
 *	psw_start	Called by midlayer to restart a device once
 *			more command openings become available.
 *
 *	psw_async	Called by midlayer when an asynchronous event
 *			from the adapter occurs.
 *
 *	psw_done	Called by the midlayer when an xfer has completed.
 */
struct scsipi_periphsw {
	int		(*psw_error)(struct scsipi_xfer *);
	void	(*psw_start)(struct scsipi_periph *);
	int		(*psw_async)(struct scsipi_periph *, scsipi_async_event_t, void *);
	void	(*psw_done)(struct scsipi_xfer *, int);
};

/*
 * The following defines the scsipi_xfer queue.
 */
TAILQ_HEAD(scsipi_xfer_queue, scsipi_xfer);

struct scsipi_channel;

/*
 * scsipi_periph:
 *
 *	This structure describes the path between a peripherial device
 *	and an adapter.  It contains a pointer to the adapter channel
 *	which in turn contains a pointer to the adapter.
 *
 * XXX Given the way NetBSD's autoconfiguration works, this is ...
 * XXX nasty.
 *
 *	Well, it's a lot nicer than it used to be, but there could
 *	still be an improvement.
 */
struct scsipi_periph {
	struct device *periph_dev;				/* pointer to peripherial's device */
	struct scsipi_channel *periph_channel;	/* channel we're connected to */

	LIST_ENTRY(scsipi_periph) periph_hash;	/* link in channel's table of periphs */

	const struct scsipi_periphsw *periph_switch; /* peripherial's entry points */
	int			periph_openings;	/* max # of outstanding commands */
	int			periph_active;		/* current # of outstanding commands */
	int			periph_sent;		/* current # of commands sent to adapt*/
	int			periph_mode;		/* operation modes, CAP bits */
	int			periph_period;		/* sync period (factor) */
	int			periph_offset;		/* sync offset */

	/*
	 * Information gleaned from the inquiry data.
	 */
	u_int8_t	periph_type;		/* basic device type */
	int			periph_cap;			/* capabilities */
	int			periph_quirks;		/* device's quirks */
	int			periph_flags;		/* misc. flags */
	int			periph_dbflags;		/* debugging flags */
	int			periph_target;		/* target ID (drive # on ATAPI) */
	int			periph_lun;			/* LUN (not used on ATAPI) */
	int			periph_version;		/* ANSI SCSI version */
	int			periph_qfreeze;		/* queue freeze count */

	/* Bitmap of free command tags. */
	u_int32_t	periph_freetags[PERIPH_NTAGWORDS];

	/* Pending scsipi_xfers on this peripherial. */
	struct scsipi_xfer_queue periph_xferq;

	struct callout periph_callout;

	/* xfer which has a pending CHECK_CONDITION */
	struct scsipi_xfer *periph_xscheck;
};

/*
 * Error values an adapter driver may return
 */
typedef enum {
	XS_NOERROR,				/* there is no error, (sense is invalid)  */
	XS_SENSE,				/* Check the returned sense for the error */
	XS_SHORTSENSE,			/* Check the ATAPI sense for the error	  */
	XS_DRIVER_STUFFUP,		/* Driver failed to perform operation     */
	XS_RESOURCE_SHORTAGE,	/* adapter resource shortage		  */
	XS_SELTIMEOUT,			/* The device timed out.. turned off?     */
	XS_TIMEOUT,				/* The Timeout reported was caught by SW  */
	XS_BUSY,				/* The device busy, try again later?      */
	XS_RESET,				/* bus was reset; possible retry command  */
	XS_REQUEUE				/* requeue this command */
} scsipi_xfer_result_t;

/*
 * Each scsipi transaction is fully described by one of these structures
 * It includes information about the source of the command and also the
 * device and adapter for which the command is destined.
 *
 * Before the HBA is given this transaction, channel_q is the linkage on
 * the related channel's chan_queue.
 *
 * When the this transaction is taken off the channel's chan_queue and
 * the HBA's request entry point is called with this transaction, the
 * HBA can use the channel_q tag for whatever it likes until it calls
 * scsipi_done for this transaction, at which time it has to stop
 * using channel_q.
 *
 * After scsipi_done is called with this transaction and if there was an
 * error on it, channel_q then becomes the linkage on the related channel's
 * chan_complete cqueue.
 *
 * The device_q member is maintained by the scsipi middle layer.  When
 * a device issues a command, the xfer is placed on that device's
 * pending commands queue.  When an xfer is done and freed, it is taken
 * off the device's queue.  This allows for a device to wait for all of
 * its pending commands to complete.
 */
struct scsipi_xfer {
	TAILQ_ENTRY(scsipi_xfer) channel_q; /* entry on channel queue */
	TAILQ_ENTRY(scsipi_xfer) device_q;  /* device's pending xfers */
	struct callout	xs_callout;			/* callout for adapter use */
	int				xs_control;			/* control flags */
	volatile int	xs_status;			/* status flags */
	struct scsipi_periph *xs_periph;	/* peripherial doing the xfer */
	int				xs_retries;			/* the number of times to retry */
	int				xs_requeuecnt;		/* number of requeues */
	int				timeout;			/* in milliseconds */
	struct scsipi_generic *cmd;			/* The scsipi command to execute */
	int				cmdlen;				/* how long it is */
	u_char			*data;				/* DMA address OR a uio address */
	int				datalen;			/* data len (blank if uio) */
	int				resid;				/* how much buffer was not touched */
	scsipi_xfer_result_t error;			/* an error value */
	struct buf		*bp;				/* If we need to associate with a buf */
	union {
		struct scsi_sense_data scsi_sense; /* 32 bytes */
		u_int32_t atapi_sense;
	} sense;

	struct scsipi_xfer *xs_sensefor;/* we are requesting sense for this xfer */

	u_int8_t		status;				/* SCSI status */

	/*
	 * Info for tagged command queueing.  This may or may not
	 * be used by a given adapter driver.  These are the same
	 * as the bytes in the tag message.
	 */
	u_int8_t		xs_tag_type;		/* tag type */
	u_int8_t		xs_tag_id;			/* tag ID */

	struct scsipi_generic cmdstore
	    __attribute__ ((aligned (4)));/* stash the command in here */
};

/*
 * scsipi_bustype:
 *
 *	This structure describes a SCSIPI bus type.
 *	The bustype_type member is shared with struct ata_bustype
 *	(because we can ata, atapi or scsi busses to the same controller)
 */
struct scsipi_bustype {
	int	bustype_type;		/* symbolic name of type */

	void	(*bustype_cmd)(struct scsipi_xfer *);
	int		(*bustype_interpret_sense)(struct scsipi_xfer *);
	void	(*bustype_printaddr)(struct scsipi_periph *);
	void	(*bustype_kill_pending)(struct scsipi_periph *);
};

/*
 * scsipi_channel:
 *
 *	This structure describes a single channel of a SCSIPI adapter.
 *	An adapter may have one or more channels.  See the comment above
 *	regarding the resource counter.
 *	Note: chan_bustype has to be first member, as its bustype_type member
 * 	is shared with the aa_bustype member of struct ata_atapi_attach.
 */

#define	SCSIPI_CHAN_PERIPH_BUCKETS	16
#define	SCSIPI_CHAN_PERIPH_HASHMASK	(SCSIPI_CHAN_PERIPH_BUCKETS - 1)

struct scsipi_channel {
	const struct scsipi_bustype *chan_bustype; /* channel's bus type */
	const char *chan_name;			/* this channel's name */

	struct scsipi_adapter *chan_adapter; /* pointer to our adapter */

	/* Periphs for this channel. */
	LIST_HEAD(, scsipi_periph) chan_periphtab[SCSIPI_CHAN_PERIPH_BUCKETS];

	int		chan_channel;		/* channel number */
	int		chan_flags;			/* channel flags */
	int		chan_openings;		/* number of command openings */
	int		chan_max_periph;	/* max openings per periph */

	int		chan_ntargets;		/* number of targets */
	int		chan_nluns;			/* number of luns */
	int		chan_id;			/* adapter's ID for this channel */

	int		chan_defquirks;		/* default device's quirks */

//	struct proc *chan_thread;	/* completion thread */
	int		chan_tflags;		/* flags for the completion thread */

	int		chan_qfreeze;		/* freeze count for queue */

	/* Job queue for this channel. */
	struct scsipi_xfer_queue chan_queue;

	/* Completed (async) jobs. */
	struct scsipi_xfer_queue chan_complete;

	/* callback we may have to call from completion thread */
	void (*chan_callback)(struct scsipi_channel *, void *);
	void *chan_callback_arg;

	/* callback we may have to call after forking the kthread */
	void (*chan_init_cb)(struct scsipi_channel *, void *);
	void *chan_init_cb_arg;
};

#endif /* _DEV_SCSIPI_SCSIPICONF_H_ */
